/*	$Id: spaceinvaders.c,v 1.70 2017/05/26 19:38:11 hacki Exp $ */

/*
 * Copyright (c) 2017 Marcus Glocker <marcus@nazgul.ch>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Another version of the famous Space Invaders arcade video game created
 * by Tomohiro Nishikado and published in 1978.  I tried to retain the
 * original game rules as far as possible.
 */

#include <curses.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include "spaceinvaders.h"

/*
 * Prototypes.
 */
void	usage(int);
int	intro(void);
void	timer_setup(void);
void	timer_clear(void);
void	keyboard_handler(void);
void	hiscore_load(void);
void	hiscore_save(void);
void	ctrl_init(int);
void	board_init(int);
void	board_draw(void);
void	board_write(int, int, const char *, ...);
void	board_write_msg(const char *, ...);
void	board_update(int);
void	board_update_player_move(int);
void	board_update_player_laser_trigger(void);
void	board_update_player_laser(void);
int	board_update_aliens_move(void);
int	board_update_aliens_missile_trigger(void);
int	board_update_aliens_missile(void);
void	board_update_mystery_move(void);
void	board_update_mystery_trigger(void);
void	board_backing_store(int);
int	lookup_sprite(char);
void	player_score(int);
void	player_destroy(void);
void	player_redraw_lives(void);
void	game_reset(int);
void	game_pause(void);
void	game_over(void);
char	getkey(void);
void	flush_stdin(void);
void	debug_setup(char *);
void	debug_write(char *, ...);

int
main(int argc, char *argv[])
{
	int ch;

	while ((ch = getopt(argc, argv, "d:hsv")) != -1) {
		switch (ch) {
		case 'd':
			/* Undocumented. */
			debug_setup(optarg);
			break;
		case 'h':
			usage(MODE_USAGE_HELP);
			/* NOTREACHED */
		case 's':
			silent_mode = 1;
			break;
		case 'v':
			usage(MODE_USAGE_VERSION);
			/* NOTREACHED */
		default:
			usage(MODE_USAGE_HELP);
			/* NOTREACHED */
		}
	}

	/* Init curses. */
	initscr();

	/* Handle signals. */
	signal(SIGALRM, board_update);
	signal(SIGINT, SIG_IGN);

	/* Make STDIN non-blocking. */
	fcntl(STDIN_FILENO, F_SETFL, fcntl(0, F_GETFL) | O_NONBLOCK);

	/* Load hi-score. */
	hiscore_load();

	/* Display intro. */
	if (intro()) {
		/* Start the game. */
		D(1, "Start wave %d", ctrl.wave);
		game_reset(MODE_CTRL_INIT_FULL);
		keyboard_handler();
		D(1, "End wave %d", ctrl.wave);
	}

	/* End curses. */
	endwin();

	return 0;
}

void
usage(int mode)
{
	if (mode == MODE_USAGE_VERSION)
		fprintf(stderr, "spaceinvaders version %s\n", VERSION);
	if (mode == MODE_USAGE_HELP) {
		fprintf(stderr, "usage: %s ", __progname);
		fprintf(stderr, "[-hsv]\n");
	}
	exit(1);
}

/*
 * RETURN codes:
 *	0 = Quit.
 *	1 = Start the game.
 */
int
intro(void)
{
	char ch = 0;

	board_init(MODE_BOARD_INIT_INTRO);
	board_write(3,  39, "PLAY");
	board_write(5,  34, "SPACE INVADERS");
	board_write(7,  30, "*SCORE ADVANCE TABLE*");
	board_write(8,  34, "T = ? Mystery");
	board_write(9,  34, "Y = 30 Points");
	board_write(10, 34, "O = 20 Points");
	board_write(11, 34, "H = 10 Points");
	board_write(13, 31, "*GAME CONTROL KEYS*");
	board_write(14, 22, "CURSOR-LEFT / H = Move Laser Base Left");
	board_write(15, 21, "CURSOR-RIGHT / L = Move Laser Base Right");
	board_write(16, 26, "CURSOR-UP / SPACE = Laser Shot");
	board_write(17, 34, "P = Pause Game");
	board_write(18, 34, "Q = Quit Game");
	board_write(21, 22, "PRESS S TO START THE GAME OR Q TO QUIT");
	board_draw();

	/* Wait for user to start game. */
	while (1) {
		ch = getkey();

		switch (ch) {
		case 's':
			return 1;
			/* NOTREACHED */
		case 'q':
			return 0;
			/* NOTREACHED */
		}
	}
}

void
timer_setup(void)
{
	struct itimerval t;

	t.it_interval.tv_sec = 0;
	t.it_interval.tv_usec = SPEED_BOARD_UPDATE;
	t.it_value.tv_sec = 0;
	t.it_value.tv_usec = 1; /* Start timer almost immediately. */
	if (setitimer(ITIMER_REAL, &t, NULL) == -1)
		err(1, "setitimer");
}

void
timer_clear(void)
{
	struct itimerval t;

	t.it_interval.tv_sec = 0;
	t.it_interval.tv_usec = 0;
	t.it_value.tv_sec = 0;
	t.it_value.tv_usec = 0;
	if (setitimer(ITIMER_REAL, &t, NULL) == -1)
		err(1, "setitimer");
}

void
keyboard_handler(void)
{
	int quit = 0;
	char ch = 0;

	while (!quit) {
		ch = getkey();

		if (ctrl.game_pause) {
			switch (ch) {
			case 'p':
				/* Handle in main switch. */
				break;
			default:
				/* No other keys accepted. */
				continue;
			}
		}

		if (ctrl.game_over) {
			switch (ch) {
			case 's':
				/* Start new game. */
				game_reset(MODE_CTRL_INIT_FULL);
				continue;
			case 'q':
				/* Handle in main switch. */
				break;
			default:
				/* No other keys accepted. */
				continue;
			}
		}

		switch (ch) {
		case 67:	/* Cursor right. */
		case 'l':
			/* Move laser base to the right. */
			board_update_player_move(DIRECTION_RIGHT);
			break;
		case 68:	/* Cursor left. */
		case 'h':
			/* Move laser base to the left. */
			board_update_player_move(DIRECTION_LEFT);
			break;
		case 65:	/* Cursor up. */
		case ' ':
			/* Laser base shots. */
			board_update_player_laser_trigger();
			break;
		case 'p':
			/* Pause or resume the game. */
			game_pause();
			break;
		case 'q':
			/* Quit the game. */
			hiscore_save();
			quit = 1;
			break;
		default:
			/* No other keys accepted. */
			break;
		}
	}
}

void
hiscore_load(void)
{
	int hiscore_fd;
	char hiscore[5];
	char filepath[1024];
	char *home;
	const char *errstr = NULL;

	home = getenv("HOME");
	snprintf(filepath, sizeof(filepath), "%s/%s", home, FILENAME_HISCORE);

	hiscore_fd = open(filepath, O_RDONLY);
	if (hiscore_fd == -1) {
		/* Can't open hiscore file. */
		ctrl.player_hiscore = 0;
		return;
	}

	memset(hiscore, 0, sizeof(hiscore));
	read(hiscore_fd, hiscore, sizeof(hiscore)-1);
#ifdef __OpenBSD__
	ctrl.player_hiscore = strtonum(hiscore, 0, SCORE_MAX, &errstr);
#else
	ctrl.player_hiscore = atoi(hiscore);
#endif
	if (errstr)
		err(1, "%s", errstr);
	close(hiscore_fd);
}

void
hiscore_save(void)
{
	int hiscore_fd;
	char hiscore[5];
	char filepath[1024];
	char *home;

	home = getenv("HOME");
	snprintf(filepath, sizeof(filepath), "%s/%s", home, FILENAME_HISCORE);

	hiscore_fd = open(filepath, O_CREAT|O_TRUNC|O_WRONLY, 0644);
	if (hiscore_fd == -1) {
		/* Can't open hiscore file. */
		return;
	}

	snprintf(hiscore, sizeof(hiscore), "%04d", ctrl.player_hiscore);
	write(hiscore_fd, hiscore, 4);
	close(hiscore_fd);
}

void
ctrl_init(int mode)
{
	if (mode == MODE_CTRL_INIT_FULL) {
		ctrl.wave = 0;
		ctrl.player_score = 0;
		ctrl.player_hiscore_new = 0;
		ctrl.player_lives = PLAYER_LIVES;
		ctrl.score_extra_live = SCORE_EXTRA_LIVE;
	}
	ctrl.player_pos_x = 14;
	ctrl.player_shot_inprogress = 0;
	ctrl.player_laser_cycle = 0;
	ctrl.player_laser_counter = 0;
	ctrl.aliens_num = ALIENS_NUM;
	ctrl.aliens_shot_inprogress = 0;
	ctrl.aliens_dir_now = DIRECTION_RIGHT;
	ctrl.aliens_dir_last = DIRECTION_RIGHT;
	ctrl.aliens_move_cycle = 0;
	ctrl.aliens_move_usec = SPEED_ALIENS_MOVE;
	ctrl.aliens_missile_trigger_cycle = 0;
	ctrl.aliens_missile_cycle = 0;
	ctrl.mystery_trigger_cycle = 0;
	ctrl.mystery_move_cycle = 0;
	ctrl.mystery_inprogress = 0;
	ctrl.mystery_laser_top = MYSTERY_LASER_TOP;
	ctrl.mystery_dir = DIRECTION_RIGHT;
	ctrl.msg_cleanup_cycle = 0;
	ctrl.game_over = 0;
	ctrl.game_pause = 0;
}

void
board_init(int mode)
{
	int y, x, i;

	clear();

	/* Initialize board with spaces. */
	for (y = 0; y < BOARD_ROWS; y++)
		memset(board[y], sprites[SPACE].graph, BOARD_COLS);

	/* Scores. */
	board_write(0, 0, "%s", txt_score);
	board_write(0, 16, "%s", txt_hiscore);
	board_write(1, 0, "%04d", ctrl.player_score);
	board_write(1, 16, "%04d", ctrl.player_hiscore);

	if (mode == MODE_BOARD_INIT_INTRO)
		return;

	/* Laser base defense. */
	for (y = 0, x = 12; y < 4; y++, x += 16)
		memset(board[17]+x, sprites[SHIELD_3].graph, 6);
	for (y = 0, x = 11; y < 4; y++, x += 16)
		memset(board[18]+x, sprites[SHIELD_3].graph, 8);
	for (y = 0, x = 11, i = 0; y < 8; y++, x += i) {
		memset(board[19]+x, sprites[SHIELD_3].graph, 2);
		if (i == 0)
			i = 6;
		else if (i == 6)
			i = 10;
		else if (i == 10)
			i = 6;
	}

	/* Aliens 1. */
	for (y = 0, x = 18; y < ALIENS_PER_LINE; y++, x += 4)
		board[6+ctrl.wave][x] = sprites[ALIEN_1].graph;

	/* Aliens 2. */
	for (y = 0, x = 18; y < ALIENS_PER_LINE; y++, x += 4)
		board[7+ctrl.wave][x] = sprites[ALIEN_2].graph;
	for (y = 0, x = 18; y < ALIENS_PER_LINE; y++, x += 4)
		board[8+ctrl.wave][x] = sprites[ALIEN_2].graph;

	/* Aliens 3. */
	for (y = 0, x = 18; y < ALIENS_PER_LINE; y++, x += 4)
		board[9+ctrl.wave][x] = sprites[ALIEN_3].graph;
	for (y = 0, x = 18; y < ALIENS_PER_LINE; y++, x += 4)
		board[10+ctrl.wave][x] = sprites[ALIEN_3].graph;

	/* Laser Base. */
	board[21][ctrl.player_pos_x] = sprites[LASER_BASE].graph;

	/* Floor. */
	memset(board[22], sprites[FLOOR].graph, BOARD_COLS);

	/* Laser base lives. */
	board_write(23, 0, "%d", ctrl.player_lives);
	for (y = 0, x = 8; y < (ctrl.player_lives - 1); y++, x += 2)
		board[23][x] = sprites[LASER_BASE].graph;

	board_draw();
}

void
board_draw(void)
{
	int y, x;

	for (y = 0; y < BOARD_ROWS; y++) {
		for (x = 0; x < BOARD_COLS; x++)
			mvaddch(y, x, board[y][x]);
	}
	move(23, 79);
	refresh();
}

void
board_write(int y, int x, const char *msg, ...)
{
	int i;
	char buf[1024];
	va_list ap;

	va_start(ap, msg);
	vsnprintf(buf, sizeof(buf), msg, ap);
	va_end(ap);

	for (i = 0; i < strlen(buf); i++, x++)
		board[y][x] = buf[i];
}

void
board_write_msg(const char *msg, ...)
{
	char buf[1024];
	va_list ap;

	va_start(ap, msg);
	vsnprintf(buf, sizeof(buf), msg, ap);
	va_end(ap);

	ctrl.msg_cleanup_cycle = 0;
	memset(board[0]+35, sprites[SPACE].graph, 39);
	if (msg == NULL) {
		/* Just wipe the message field. */
		return;
	}

	board_write(0, 35, "%s", buf);
}

void
board_update(int sig)
{
	int r1 = 0;
	int r2 = 0;
	int r3 = 0;

	if (sig != SIGALRM) {
		/* We just care about the timer signal. */
		return;
	}
	if (lock_signal) {
		/* Signal handler got locked, try again later. */
		D(1, "Signal handler is locked");
		return;
	}
	lock_signal = 1;	/* Lock ourself just to be sure. */

	/* Update cycle timers. */
	ctrl.aliens_move_cycle += SPEED_BOARD_UPDATE;
	ctrl.aliens_missile_cycle += SPEED_BOARD_UPDATE;
	ctrl.aliens_missile_trigger_cycle += SPEED_BOARD_UPDATE;
	ctrl.player_laser_cycle += SPEED_BOARD_UPDATE;
	ctrl.mystery_trigger_cycle += SPEED_BOARD_UPDATE;
	ctrl.mystery_move_cycle += SPEED_BOARD_UPDATE;
	ctrl.msg_cleanup_cycle += SPEED_BOARD_UPDATE;

	if (ctrl.aliens_move_cycle >= ctrl.aliens_move_usec) {
		/* Update aliens movement. */
		ctrl.aliens_move_cycle = 0;
		r1 = board_update_aliens_move();
		if (r1 == 1) {
			/* Change moving direction. */
			r1 = board_update_aliens_move();
		}
		if (r1 == 2) {
			/* Aliens land on the floor. */
			game_over();
			lock_signal = 0;
			return;
		}
	}
	if (ctrl.aliens_missile_cycle == SPEED_ALIENS_MISSILE_MOVE) {
		ctrl.aliens_missile_cycle = 0;
		/* Update aliens missile. */
		r2 = board_update_aliens_missile();
	}
	if (ctrl.aliens_missile_trigger_cycle == SPEED_ALIENS_MISSILE_TRIGGER) {
		/* Trigger alien missile. */
		ctrl.aliens_missile_trigger_cycle = 0;
		r3 = board_update_aliens_missile_trigger();
	}
	if (ctrl.player_laser_cycle == SPEED_LASER_SHOT_MOVE) {
		/* Update players laser. */
		ctrl.player_laser_cycle = 0;
		board_update_player_laser();
	}
	if (ctrl.mystery_trigger_cycle == SPEED_MYSTERY_TRIGGER) {
		/* Trigger mystery ship to appear. */
		ctrl.mystery_trigger_cycle = 0;
		board_update_mystery_trigger();
	}
	if (ctrl.mystery_move_cycle == SPEED_MYSTERY_MOVE) {
		/* Update mystery ship movement. */
		ctrl.mystery_move_cycle = 0;
		board_update_mystery_move();
	}
	if (ctrl.msg_cleanup_cycle == SPEED_TEXT_CLEANUP) {
		/* Cleanup text messages. */
		ctrl.msg_cleanup_cycle = 0;
		memset(board[0]+35, sprites[SPACE].graph, 39);
	}

	if (ctrl.aliens_move_cycle == 0 ||
	    ctrl.aliens_missile_cycle == 0 ||
	    ctrl.aliens_missile_trigger_cycle == 0 ||
	    ctrl.player_laser_cycle == 0 ||
	    ctrl.mystery_trigger_cycle == 0 ||
	    ctrl.mystery_move_cycle == 0 ||
	    ctrl.msg_cleanup_cycle == 0) {
		/* Refresh screen if update is pending. */
		board_draw();
	}

	/* Make a quick break when the player gets destroyed. */
	if (r2 || r3)
		player_destroy();

	lock_signal = 0;
}

void
board_update_player_move(int action)
{
	int old_player_pos_x = ctrl.player_pos_x;

	if (action == DIRECTION_RIGHT) {
		if (ctrl.player_pos_x == (BOARD_COLS-1)) {
			/* Laser base touches right border. */
			return;
		}
		ctrl.player_pos_x++;
	}
	if (action == DIRECTION_LEFT) {
		if (ctrl.player_pos_x == 0) {
			/* Laser base touches the left border. */
			return;
		}
		ctrl.player_pos_x--;
	}

	/* Clear old position. */
	board[21][old_player_pos_x] = sprites[SPACE].graph;

	/* Set new position. */
	board[21][ctrl.player_pos_x] = sprites[LASER_BASE].graph;

	board_draw();
}

void
board_update_player_laser_trigger(void)
{
	int sprite_next, shot_valid = 0;

	if (ctrl.player_shot_inprogress) {
		/* Another shot is still in progress. */
		return;
	}

	sprite_next = lookup_sprite(board[20][ctrl.player_pos_x]);

	if (sprite_next == SPACE) {
		/* The way is free to place the laser shot. */
		board[20][ctrl.player_pos_x] = sprites[LASER_SHOT].graph;
		ctrl.player_shot_inprogress++;
		shot_valid = 1;
	}

	if (sprite_next == ALIEN_1 ||
	    sprite_next == ALIEN_2 ||
	    sprite_next == ALIEN_3) {
		/* Laser base destroys alien right above. */
		D(1, "Laser base destroys alien right above");
		board[20][ctrl.player_pos_x] = sprites[SPACE].graph;
		player_score(sprite_next);
		shot_valid = 1;
	}

	if (shot_valid) {
		/* Laser shot was successful. */
		ctrl.player_laser_counter++;
		if (ctrl.player_laser_counter > ctrl.mystery_laser_top)
			ctrl.player_laser_counter = 1;
		board_draw();
	}
}

void
board_update_player_laser(void)
{
	int y, x, sprite, sprite_next;
	int defense_damage;

	for (y = 3; y < (BOARD_ROWS-2); y++) {
		for (x = 0; x < BOARD_COLS; x++) {
			sprite = lookup_sprite(board[y][x]);
			if (sprite != LASER_SHOT)
				continue;

			if ((y-1) == 2) {
				/* Laser base hits nothing. */
				board[y][x] = sprites[SPACE].graph;
				ctrl.player_shot_inprogress--;
				continue;
			}

			sprite_next = lookup_sprite(board[y-1][x]);

			if (sprite_next == SPACE) {
				/* The way is free. */
				board[y][x] = sprites[SPACE].graph;
				board[y-1][x] = sprites[LASER_SHOT].graph;
			}

			if (sprite_next == ALIEN_1_MISSILE ||
			    sprite_next == ALIEN_2_MISSILE ||
			    sprite_next == ALIEN_3_MISSILE) {
				switch (arc4random_uniform(2)) {
				case 0:
					/* Laser shot passes alien missile. */
					D(1, "Laser shot pass alien missile");
					board[y][x] =
					    sprites[sprite_next].graph;
					board[y-1][x] = sprites[sprite].graph;
					break;
				case 1:
					/* Laser shot destroys alien missile. */
					D(1, "Laser shot kill alien missile");
					board[y][x] = sprites[SPACE].graph;
					board[y-1][x] = sprites[sprite].graph;
					ctrl.aliens_shot_inprogress--;
					break;
				}
			}

			if (sprite_next == ALIEN_1 ||
			    sprite_next == ALIEN_2 ||
			    sprite_next == ALIEN_3) {
				/* Laser base destroys alien. */
				D(1, "Laser base destroys alien");
				board[y][x] = sprites[SPACE].graph;
				board[y-1][x] = sprites[SPACE].graph;
				ctrl.player_shot_inprogress--;
				player_score(sprite_next);
			}

			if (sprite_next == MYSTERY_SHIP) {
				/* Laser base destroys mystery ship. */
				D(1, "Laser base destroys mystery ship");
				board[y][x] = sprites[SPACE].graph;
				board[y-1][x] = sprites[SPACE].graph;
				ctrl.player_shot_inprogress--;
				ctrl.mystery_inprogress--;
				player_score(sprite_next);
			}

			if (sprite_next == SHIELD_3 ||
			    sprite_next == SHIELD_2 ||
			    sprite_next == SHIELD_1) {
				/* Laser base damages shield. */
				defense_damage =
				    sprites[sprite_next].protection -
				    sprites[sprite].damage;
				switch (defense_damage) {
				case 2:
					/* Shield damaged. */
					board[y-1][x] =
					    sprites[SHIELD_2].graph;
					break;
				case 1:
					/* Shield damaged more. */
					board[y-1][x] =
					    sprites[SHIELD_1].graph;
					break;
				default:
					/* Shield cracked. */
					board[y-1][x] =
					    sprites[SPACE].graph;
					break;
				}
				board[y][x] = sprites[SPACE].graph;
				ctrl.player_shot_inprogress--;
			}
		}
	}
}

/*
 * RETURN codes:
 *	0 = Regular move done.
 *	1 = Aliens touches border and need to change the direction.
 *	2 = Aliens lands on floor which means the game is over.
 */
int
board_update_aliens_move(void)
{
	int y, x, sprite, sprite_next;

	board_backing_store(BBS_BACKUP);
	board_backing_store(BBS_REMOVE_ALIENS);

	for (y = 3; y < (BOARD_ROWS-2); y++) {
		for (x = 0; x < BOARD_COLS; x++) {
			sprite = lookup_sprite(board[y][x]);
			if (sprite != ALIEN_1 &&
			    sprite != ALIEN_2 &&
			    sprite != ALIEN_3)
				continue;

			if ((y+1) == 22) {
				/* Aliens land on the floor. */
				D(1, "Aliens land on the floor");
				return 2;
			}

			if (ctrl.aliens_dir_now == DIRECTION_RIGHT) {
				if ((x+1) == BOARD_COLS) {
					/* Alien touches the right border. */
					ctrl.aliens_dir_now = DIRECTION_DOWN;
					return 1;
				}

				sprite_next = lookup_sprite(board[y][x+1]);
				if (sprite_next == LASER_SHOT) {
					/* Laser shot gets eliminated. */
					ctrl.player_shot_inprogress--;
				}

				board_bs[y][x+1] = board[y][x];
			}

			if (ctrl.aliens_dir_now == DIRECTION_LEFT) {
				if ((x-1) < 0) {
					/* Alien touches the left border. */
					ctrl.aliens_dir_now = DIRECTION_DOWN;
					return 1;
				}

				sprite_next = lookup_sprite(board[y][x-1]);
				if (sprite_next == LASER_SHOT) {
					/* Laser shot gets eliminated. */
					ctrl.player_shot_inprogress--;
				}

				board_bs[y][x-1] = board[y][x];
			}

			if (ctrl.aliens_dir_now == DIRECTION_DOWN) {
				sprite_next = lookup_sprite(board_bs[y+1][x]);
				if (sprite_next == LASER_SHOT) {
					/* Laser base destroys alien. */
					ctrl.player_shot_inprogress--;
					board[y][x] = sprites[SPACE].graph;
					player_score(sprite);
				}

				board_bs[y+1][x] = board[y][x];
			}
		}
	}

	if (ctrl.aliens_dir_now == DIRECTION_DOWN) {
		/* Change moving direction. */ 
		if (ctrl.aliens_dir_last == DIRECTION_RIGHT) {
			ctrl.aliens_dir_now = DIRECTION_LEFT;
			ctrl.aliens_dir_last = DIRECTION_LEFT;
		} else if (ctrl.aliens_dir_last == DIRECTION_LEFT) {
			ctrl.aliens_dir_now = DIRECTION_RIGHT;
			ctrl.aliens_dir_last = DIRECTION_RIGHT;
		}
	}

	board_backing_store(BBS_RESTORE);

	return 0;
}

/*
 * RETURN codes:
 *	1 = Alien destroys player.
 */
int
board_update_aliens_missile_trigger(void)
{
	int y, x, i, loops, sprite, sprite_next;
	int aliens_num; /* Number of aliens who could shot. */
	int alien_shots; /* Alien we've choosen to shot. */
	struct position {
		int sprite;
		int sprite_next;
		int y;
		int x;
		int did_shot;
	} pos[ALIENS_PER_LINE];

	D(2, "aliens_shot_inprogress=%d", ctrl.aliens_shot_inprogress);
	if (ctrl.aliens_shot_inprogress == ALIENS_SHOT_MAX) {
		/* Aliens maximum parallel shots reached. */
		D(1, "Aliens maximum parallel shots reached");
		return 0;
	}

	if (arc4random_uniform(2)) {
		/* Aliens don't always shot. */
		return 0;
	}

	aliens_num = 0;
	memset(&pos, 0, sizeof(struct position));

	for (y = 3; y < (BOARD_ROWS-2); y++) {
		for (x = 0; x < BOARD_COLS; x++) {
			sprite = lookup_sprite(board[y][x]);
			if (sprite != ALIEN_1 &&
			    sprite != ALIEN_2 &&
			    sprite != ALIEN_3)
				continue;

			sprite_next = lookup_sprite(board[y+1][x]);
			if (sprite_next == SPACE ||
			    sprite_next == LASER_BASE) {
				pos[aliens_num].sprite = sprite;
				pos[aliens_num].sprite_next = sprite_next;
				pos[aliens_num].y = y+1;
				pos[aliens_num].x = x;
				aliens_num++;
			}
		}
	}
	if (aliens_num == 0) {
		/* No alien can shot currently. */
		return 0;
	}

	loops = ALIENS_SHOT_MAX - ctrl.aliens_shot_inprogress;
	for (i = 0; i < loops; i++) {
		alien_shots = arc4random_uniform(aliens_num);
		if (pos[alien_shots].did_shot) {
			/* Alien already did shot in this round. */
			continue;
		}
		pos[alien_shots].did_shot = 1;
		y = pos[alien_shots].y;
		x = pos[alien_shots].x;
		if (pos[alien_shots].sprite_next == LASER_BASE) {
			/* Alien destroys player right below. */
			D(1, "Alien destroys player right below");
			board[y][x] = sprites[EXPLOSION].graph;
			return 1;
		}
		if (pos[alien_shots].sprite == ALIEN_1)
			board[y][x] = sprites[ALIEN_1_MISSILE].graph;
		else if (pos[alien_shots].sprite == ALIEN_2)
			board[y][x] = sprites[ALIEN_2_MISSILE].graph;
		else if (pos[alien_shots].sprite == ALIEN_3)
			board[y][x] = sprites[ALIEN_3_MISSILE].graph;
		ctrl.aliens_shot_inprogress++;
	}

	return 0;
}

/*
 * RETURN codes:
 *	1 = Alien destroys laser base.
 */
int
board_update_aliens_missile(void)
{
	int r = 0;
	int y, x, sprite, sprite_next;
	int defense_damage;

	board_backing_store(BBS_BACKUP);
	board_backing_store(BBS_REMOVE_ALIENS_MISSILE);

	for (y = 3; y < (BOARD_ROWS-2); y++) {
		for (x = 0; x < BOARD_COLS; x++) {
			sprite = lookup_sprite(board[y][x]);
			if (sprite != ALIEN_1_MISSILE &&
			    sprite != ALIEN_2_MISSILE &&
			    sprite != ALIEN_3_MISSILE)
				continue;

			if (y == 21) {
				/* Alien hits nothing. */
				board_bs[y][x] = sprites[SPACE].graph;
				ctrl.aliens_shot_inprogress--;
				continue;
			}

			sprite_next = lookup_sprite(board[y+1][x]);

			if (sprite_next == SPACE) {
				/* The way is free. */
				board_bs[y+1][x] = sprites[sprite].graph;
			}

			if (sprite_next == LASER_SHOT) {
				switch (arc4random_uniform(2)) {
				case 0:
					/* Alien missile passes laser shot. */
					D(1, "Alien missile passes laser shot");
					board_bs[y][x] =
					    sprites[sprite_next].graph;
					board_bs[y+1][x] =
					    sprites[sprite].graph;
					break;
				case 1:
					/* Alien shot destroys laser shot. */
					D(1, "Alien shot destroys laser shot");
					board_bs[y][x] = sprites[SPACE].graph;
					board_bs[y+1][x] =
					    sprites[sprite].graph;
					ctrl.player_shot_inprogress--;
					break;
				}
			}

			if (sprite_next == SHIELD_3 ||
			    sprite_next == SHIELD_2 ||
			    sprite_next == SHIELD_1) {
				/* Alien damages shield. */
				defense_damage = 
				    sprites[sprite_next].protection -
				    sprites[sprite].damage;
				switch (defense_damage) {
				case 2:
					/* Shield damaged. */
					board_bs[y+1][x] =
					    sprites[SHIELD_2].graph;
					break;
				case 1:
					/* Shield damaged more. */
					board_bs[y+1][x] =
					    sprites[SHIELD_1].graph;
					break;
				default:
					/* Shield cracked. */
					board_bs[y+1][x] =
					    sprites[SPACE].graph;
					break;
				}
				board_bs[y][x] = sprites[SPACE].graph;
				ctrl.aliens_shot_inprogress--;
			}

			if (sprite_next == LASER_BASE) {
				/* Alien destroys laser base. */
				D(1, "Alien destroys laser base");
				ctrl.aliens_shot_inprogress--;
				board_bs[y+1][x] = sprites[EXPLOSION].graph;
				r = 1;
			}
		}
	}

	board_backing_store(BBS_RESTORE);

	return r;
}

void
board_update_mystery_move(void)
{
	int x, sprite, sprite_next;

	for (x = 0; x < BOARD_COLS; x++) {
		sprite = lookup_sprite(board[3][x]);
		if (sprite != MYSTERY_SHIP)
			continue;

		if ((ctrl.mystery_dir == DIRECTION_LEFT  && (x-1) == -1) ||
		    (ctrl.mystery_dir == DIRECTION_RIGHT && (x+1) == 80)) {
			/* Mystery ship leaves screen. */
			D(1, "Mystery ship leaves screen");
			board[3][x] = sprites[SPACE].graph;
			ctrl.mystery_inprogress--;
			break;
		}

		if (ctrl.mystery_dir == DIRECTION_LEFT) {
			sprite_next = lookup_sprite(board[3][x-1]);
			board[3][x-1] = sprites[sprite].graph;
		} else {
			sprite_next = lookup_sprite(board[3][x+1]);
			board[3][x+1] = sprites[sprite].graph;
		}

		if (sprite_next == LASER_SHOT) {
			/* Laser shot gets eliminated. */
			ctrl.player_shot_inprogress--;
		}

		board[3][x] = sprites[SPACE].graph;
		break;
	}
}

void
board_update_mystery_trigger(void)
{
	if (ctrl.mystery_inprogress) {
		/* Mystery ship already entered. */
		return;
	}

	if (arc4random_uniform(32) == 4) {
		/* Mystery ship enters screen. */
		D(1, "Mystery ship enters screen");
		ctrl.mystery_dir = arc4random_uniform(2);
		if (ctrl.mystery_dir == DIRECTION_LEFT)
			board[3][79] = sprites[MYSTERY_SHIP].graph;
		else
			board[3][0] = sprites[MYSTERY_SHIP].graph;
		ctrl.mystery_inprogress++;
	}
}

void
board_backing_store(int action)
{
	int y, x, sprite;

	if (action == BBS_BACKUP) {
		for (y = 0; y < BOARD_ROWS; y++)
			memcpy(board_bs[y], board[y], BOARD_COLS);
	}

	if (action == BBS_RESTORE) {
		for (y = 0; y < BOARD_ROWS; y++)
			memcpy(board[y], board_bs[y], BOARD_COLS);
	}

	if (action == BBS_REMOVE_ALIENS) {
		for (y = 3; y < (BOARD_ROWS-2); y++) {
			for (x = 0; x < BOARD_COLS; x++) {
				sprite = lookup_sprite(board_bs[y][x]);
				if (sprite == ALIEN_1 ||
				    sprite == ALIEN_2 ||
				    sprite == ALIEN_3)
					board_bs[y][x] = sprites[SPACE].graph;
			}
		}
	}

	if (action == BBS_REMOVE_ALIENS_MISSILE) {
		for (y = 3; y < (BOARD_ROWS-2); y++) {
			for (x = 0; x < BOARD_COLS; x++) {
				sprite = lookup_sprite(board_bs[y][x]);
				if (sprite == ALIEN_1_MISSILE ||
				    sprite == ALIEN_2_MISSILE ||
				    sprite == ALIEN_3_MISSILE)
					board_bs[y][x] = sprites[SPACE].graph;
			}
		}
	}
}

/*
 * RETURN codes:
 *	-1 = Sprite not found.
 *	<sprite> = Successful.
 */
int
lookup_sprite(char graph)
{
	int i;

	for (i = 0; i < SPRITES_NUM; i++) {
		if (graph == sprites[i].graph)
			return i;
	}

	return -1;
}

void
player_score(int sprite)
{
	int n, score;

	if (silent_mode == 0)
		beep();

	/* Update score. */
	if (sprite == MYSTERY_SHIP) {
		if (ctrl.player_laser_counter == ctrl.mystery_laser_top) {
			/* Mystery ship gets 300 points when destroyed
			 * first time by the 23rd laser shot and afterwards
			 * by the 15th laser shot.
			 */
			if (ctrl.mystery_laser_top == MYSTERY_LASER_TOP)
				ctrl.mystery_laser_top = (MYSTERY_LASER_TOP-8);
			score = (sprites[sprite].score * 6);
		} else {
			/* Mystery ship gets random points of
			 * 50, 100, or 150.
			 */
			while ((n = arc4random_uniform(4)) == 0);
			score = (sprites[sprite].score * n);
		}
		board_write_msg("%s %d", msg_mystery, score);
	} else {
		score = sprites[sprite].score;
		ctrl.aliens_num--;
		/* Make aliens faster.
		 *
		 * History comment (source Wikipedia):
		 * While programming the game, Nishikado discovered that the
		 * processor was able to render the alien graphics faster the
		 * fewer were on screen.  Rather than design the game to
		 * compensate for the speed increase, he decided to keep it
		 * as a challenging gameplay mechanism.
		 */
		if (ctrl.aliens_num > 1) {
			ctrl.aliens_move_usec -= SPEED_ALIENS_MOVE_DEC;
		} else {
			/* Last alien gets even faster. */
			ctrl.aliens_move_usec = SPEED_ALIENS_MOVE_DEC * 4;
		}
	}
	if ((ctrl.player_score + score) <= SCORE_MAX) {
		ctrl.player_score += score;
		board_write(1, 0, "%04d", ctrl.player_score);
	}
	if (ctrl.player_score > ctrl.player_hiscore) {
		ctrl.player_hiscore = ctrl.player_score;
		board_write(1, 16, "%04d", ctrl.player_hiscore);
		if (ctrl.player_hiscore_new == 0) {
			/* Tell player about new hiscore reached. */
			ctrl.player_hiscore_new++;
			board_write_msg("%s", msg_hiscore_new);
		}
	}

	if (ctrl.player_score >= ctrl.score_extra_live) {
		/* Player gets one extra live when certain score is reached. */
		ctrl.player_lives++;
		ctrl.score_extra_live = SCORE_MAX+1;
		board_write_msg("%s", msg_extra_live);
		player_redraw_lives();
	}

	if (ctrl.aliens_num == 0) {
		/* Player wins game. */
		timer_clear();
		board_write_msg("%s", msg_player_wins);
		if (ctrl.wave == (WAVES-1)) {
			/* Start over at first wave. */
			ctrl.wave = 0;
		} else {
			/* Next wave. */
			ctrl.wave++;
			D(1, "Start wave %d", ctrl.wave);
		}
		board_draw();
		sleep(PAUSE_SEC);
		game_reset(MODE_CTRL_INIT_PARTIAL);
	}
}

void
player_destroy(void)
{
	if (silent_mode == 0)
		beep();
	timer_clear();

	/* Decrease player lives. */
	ctrl.player_lives--;
	player_redraw_lives();

	if (ctrl.player_lives == 0) {
		/* Player has no more lives. */
		game_over();
		return;
	}

	/* Reset player base to start position. */
	board[21][ctrl.player_pos_x] = sprites[SPACE].graph;
	ctrl.player_pos_x = 14;
	board[21][ctrl.player_pos_x] = sprites[LASER_BASE].graph;

	/* Have a break before restart. */
	sleep(PAUSE_SEC);
	flush_stdin();
	timer_setup();
}

void
player_redraw_lives(void)
{
	int y, x;

	for (x = 0; x < BOARD_COLS; x++)
		board[23][x] = sprites[SPACE].graph;

	board_write(23, 0, "%d", ctrl.player_lives);
	for (y = 0, x = 8; y < (ctrl.player_lives - 1); y++, x += 2)
		board[23][x] = sprites[LASER_BASE].graph;
}

void
game_reset(int mode)
{
	flush_stdin();
	ctrl_init(mode);
	board_init(MODE_BOARD_INIT_FULL);
	timer_setup();
}

void
game_pause(void)
{
	if (ctrl.game_pause == 0) {
		timer_clear();
		ctrl.game_pause = 1;
		board_write_msg("%s", msg_game_pause);
		board_draw();
	} else {
		ctrl.game_pause = 0;
		board_write_msg(NULL);
		timer_setup();
	}
}

void
game_over(void)
{
	timer_clear();
	ctrl.game_over = 1;
	board_write_msg("%s", msg_game_over);
	board_draw();
}

/*
 * RETURN codes:
 *	0 = Some error happened.
 *	<character> = Successful.
 */
char
getkey(void)
{
	int nfds;
	char ch;
	struct pollfd fdpoll[1];

	fdpoll[0].fd = STDIN_FILENO;
	fdpoll[0].events = POLLIN;

	while (1) {
		lock_signal = 0;
		nfds = poll(fdpoll, 1, -1);
		lock_signal = 1;
		if (nfds == -1) {
			if (errno == EINTR)
				continue;
			break;
		}
		if (nfds == 0)
			continue;
		if (fdpoll[0].revents & POLLIN) {
			read(STDIN_FILENO, &ch, 1);
			return ch;
		}
	}

	return 0;
}

void
flush_stdin(void)
{
	char ch;

	while (read(STDIN_FILENO, &ch, 1) != -1);
}

void
debug_setup(char *dl)
{
	const char *errstr = NULL;

	debug_level = strtonum(dl, 1, 9, &errstr);
	if (errstr != NULL)
		errx(1, "Debug level is %s: %s", errstr, dl);

	debug_fd = open(FILENAME_DEBUG, O_CREAT|O_APPEND|O_WRONLY, 0644);
	if (debug_fd == -1)
		err(1, "open");
}

void
debug_write(char *msg, ...)
{
	char buf[1024];
	int len;
	va_list ap;

	if (!debug_fd)
		return;

	va_start(ap, msg);
	vsnprintf(buf, sizeof(buf), msg, ap);
	va_end(ap);

	len = strlen(buf);
	buf[len] = '\n';
	write(debug_fd, buf, len+1);
}
